// This module implements two versions of a barrier
// 1) using conditional and atomic variables
// 2) using atomic variables + busy waiting
//
// To use one of the versions, uncomment one of the typedefs at the end of this file.

#include <atomic>
#include <mutex>
#include <condition_variable>

using namespace std;


// Barrier implemented using conditional and atomic variables.
class BarrierCondAndAtomic {
private:
    atomic<int> mFreeThreadSlots;
    long mPhaseCounter;
    int mNumThreads;
    mutex mMutex;
    condition_variable mCondVar;

public:
    BarrierCondAndAtomic(int numThreads)
            : mFreeThreadSlots(numThreads), mPhaseCounter(0), mNumThreads(numThreads), mMutex() {}

    void wait() {
        auto currentPhase = mPhaseCounter;
        if (--mFreeThreadSlots == 0) {
            // Reset.
            mFreeThreadSlots = mNumThreads;
            {
                // Locking is necessary, because if some thread is currently inside mCondVar predicate, notify_all()
                // will not signal it again (it may have old mPhaseCounter value).
                unique_lock<mutex> lock(mMutex);
                mPhaseCounter++;
            }

            mCondVar.notify_all();

            return;
        }
        else {
            unique_lock<mutex> myLock(mMutex);
            mCondVar.wait(myLock, [&currentPhase, this]{
                return currentPhase < mPhaseCounter;
            });
        }
    }
};

// Barrier implemented using atomic variables + busy waiting.
class BarrierAtomicAndBusyWaiting {
private:
    atomic<int> mFreeThreadSlots;
    atomic<int> mPhaseCounter;
    int mNumThreads;

public:
    BarrierAtomicAndBusyWaiting(int numThreads)
            : mFreeThreadSlots(numThreads), mPhaseCounter(0), mNumThreads(numThreads) {}

    void wait() {
        auto currentPhase = mPhaseCounter.load();
        if (--mFreeThreadSlots == 0) {
            // Reset.
            mFreeThreadSlots = mNumThreads;
            mPhaseCounter++;
        } else {
            // Busy waiting until the last thread resets the barrier.
            while (currentPhase == mPhaseCounter) {
                ;
            }
        }
    }
};

typedef BarrierCondAndAtomic Barrier;
//typedef BarrierAtomicAndBusyWaiting Barrier;
