#include <vector>
#include <thread>
#include <iostream>
#include <algorithm>

#include "Barrier.hpp"

using namespace std;

class Bucket {
public:
    int mSize;
    vector<int>::iterator mStart;
    vector<int>::iterator mEnd;    // Points one more value to the right out of the bucket.

    Bucket(int size, vector<int>::iterator start, vector<int>::iterator end)
            : mSize(size), mStart(start), mEnd(end) { }
};

void mergeBuckets(Bucket &left, Bucket &right) {
    vector<int> buffer(left.mSize + right.mSize, 0);
    auto leftIt = left.mStart;
    auto rightIt = right.mStart;
    auto bufferIt = buffer.begin();

    while (leftIt != left.mEnd && rightIt != right.mEnd) {
        if (*leftIt <= *rightIt) {
            *bufferIt = *leftIt;
            leftIt++;
        }
        else {
            *bufferIt = *rightIt;
            rightIt++;
        }

        bufferIt++;
    }

    while (leftIt != left.mEnd) {
        *bufferIt = *leftIt;
        bufferIt++;
        leftIt++;
    }

    while (rightIt != right.mEnd) {
        *bufferIt = *rightIt;
        bufferIt++;
        rightIt++;
    }

    bufferIt = buffer.begin();
    for (auto valuesIt = left.mStart; valuesIt != right.mEnd; valuesIt++) {
        *valuesIt = *bufferIt;
        bufferIt++;
    }
}

void parallelOddEvenSort(int workerIndex, int numWorkers, vector<Bucket> &buckets, Barrier &barrier) {
    // Sort the values in buckets first.
    sort(buckets[2 * workerIndex].mStart, buckets[2 * workerIndex].mEnd);
    sort(buckets[2 * workerIndex + 1].mStart, buckets[2 * workerIndex + 1].mEnd);
    barrier.wait();

    for (int iteration = 0; iteration < buckets.size(); iteration++) {
        // Odd phase.
        {
            int leftBucketIndex = 2 * workerIndex;
            int rightBucketIndex = 2 * workerIndex + 1;
            mergeBuckets(buckets[leftBucketIndex], buckets[rightBucketIndex]);
        }

        barrier.wait();

        // Even phase.
        if ((workerIndex + 1) < numWorkers) {
            int leftBucketIndex = 1 + 2 * workerIndex;
            int rightBucketIndex = 1 + 2 * workerIndex + 1;
            mergeBuckets(buckets[leftBucketIndex], buckets[rightBucketIndex]);
        }

        barrier.wait();
    }
}

int main() {
    // Array to sort (ascending).
    vector<int> values = { 999, 859, 394, 500, 225, 971, 958, 995, 71, 662, 598, 51, 744, 352, 514, 1 };

    int numWorkers = thread::hardware_concurrency();

    // Create buckets.
    int numBuckets = numWorkers * 2;
    int firstBucketsSize = values.size() / numBuckets;
    int lastBucketSize = values.size() - firstBucketsSize * (numBuckets - 1);
    vector<Bucket> buckets;
    for (int bucketIndex = 0; bucketIndex < numBuckets; bucketIndex++) {
        if ((bucketIndex + 1) < numBuckets) {
            buckets.emplace_back(
                    firstBucketsSize,
                    next(values.begin(), bucketIndex * firstBucketsSize),
                    next(values.begin(), (bucketIndex + 1) * firstBucketsSize)
            );
        }
        else {
            buckets.emplace_back(
                    lastBucketSize,
                    next(values.begin(), bucketIndex * firstBucketsSize),
                    values.end()
            );
        }
    }

    // Create threads.
    vector<thread> workers;
    Barrier barrier(numWorkers);
    for (int workerIndex = 0; workerIndex < numWorkers; workerIndex++) {
        workers.emplace_back(parallelOddEvenSort, workerIndex, numWorkers, ref(buckets), ref(barrier));
    }

    for (auto &worker : workers) {
        worker.join();
    }

    // Check if the values array is sorted correctly.
    for (auto it = values.begin(); next(it) != values.end(); it++) {
        if (*it > *next(it)) {
            cout << "Bug: sorted incorrectly!" << endl;
            return 1;
        }
    }

    return 0;
}
